;; Generated package description from nftables-mode.el  -*- no-byte-compile: t -*-
(define-package "nftables-mode" "1.1.0.20221221.80909" "Major mode for editing nftables" '((emacs "25.1")) :commit "94e3b544a145e05ed3af26fcde6fef317451200a" :url "https://elpa.gnu.org/packages/nftables-mode.html" :authors '(("Trent W. Buck" . "trentbuck@gmail.com")) :maintainer '(nil . "emacs-devel@gnu.org") :keywords '("convenience"))
